"""App with given blocks - is all that you need to start you application in most cases."""

from typing import Optional, Sequence, Type

from blocks.graph import Graph
from blocks.runners import AsyncRunner, Runner
from blocks.types import Block, Event
from blocks.validation import validate_blocks


class App(object):
    """
    Represents the main entrypoint for predefined sequence of blocks.

    Builds computational graph and provides interface to actually start execution.

    Example::

      >>> from blocks import App

      >>> blocks = [...]

      >>> App(blocks).run()
    """
    def __init__(
        self,
        blocks: Sequence[Block],
        terminal_event: Optional[Type[Event]] = None,
    ) -> None:
        """
        Init application instance.

        :param blocks:          Sources and processors to be included in application graph.
        :param terminal_event:  Special event which simply stops execution, when processed.
        """
        validate_blocks(blocks)
        self._graph = Graph(blocks)
        self._terminal_event = terminal_event

    def run(self, *, min_interval: float = 0.0, once: bool = False) -> None:
        """
        Start execution.

        :param min_interval:    Minimal timeout (seconds) between repetition of the whole computational sequence.
        :param once:            If True, executes the whole computational sequence only once, otherwise won't stop until
                                specific conditions (such as terminal event) will occur.
        """
        Runner(self._graph, self._terminal_event).run(interval=min_interval, once=once)

    async def run_async(self, *, min_interval: float = 0.0, once: bool = False) -> None:
        """
        Run awaitable execution.

        :param min_interval:    Minimal timeout (seconds) between repetition of the whole computational sequence.
        :param once:            If True, executes the whole computational sequence only once, otherwise won't stop until
                                specific conditions (such as terminal event) will occur.
        """
        await AsyncRunner(self._graph, self._terminal_event).run(interval=min_interval, once=once)



"""Runners are actually runtime for statically built graph."""

import asyncio
import functools
import multiprocessing as mp
import time
import traceback
from collections import deque
from typing import List, Type, Deque, Optional, Awaitable, DefaultDict, cast, Union

from blocks.graph import Graph
from blocks.logger import logger
from blocks.types import Event, Source, Processor, AsyncSource, EventOrEvents, AsyncProcessor, ParallelEvent
from blocks.sources.parallel_processor import run_parallel_processor


SyncProcessors = DefaultDict[Type[Event], List[Processor]]


class Runner(object):
    """
    Allows to actually run graph.

    - Checks sources for new events,
    - handles internal queue,
    - passes events to a specific processors,
    - check conditions for termination
    - proceeds actual shutdown of the app.
    """

    def __init__(self, graph: Graph, terminal_event: Optional[Type[Event]]) -> None:
        """
        Init runner instance.

        :param graph:           Computational graph to execute.
        :param terminal_event:  Special event which simply stops execution, when processed.
        """
        if graph.contains_async_blocks:
            raise RuntimeError('Blocks graph contains async blocks, must be run by AsyncRunner')

        self._sources = cast(List[Source], graph.sources)
        self._processors = cast(SyncProcessors, graph.processors)
        self._q: Deque[Event] = deque()
        self._alive = True
        self._pool: mp.Pool = mp.Pool(graph.count_of_parallel_tasks) if graph.count_of_parallel_tasks > 0 else None
        self._terminal_event = terminal_event

    def run(self, interval: float, once: bool) -> None:
        """
        Start execution.

        :param interval:        Minimal timeout (seconds) between repetition of the whole computational sequence.
        :param once:            If True, executes the whole computational sequence only once, otherwise won't stop until
                                specific conditions (such as terminal event) will occur.
        """
        self._tick()
        if not once:
            while self._alive:
                start = time.perf_counter()
                self._tick()
                delta = time.perf_counter() - start - interval
                logger.debug('Delta to sleep: {0}'.format(delta))
                if delta < 0:
                    time.sleep(abs(delta))
            logger.warning('Not alive anymore. Closing all resources...')
        self._close_resources()
        if not once:
            logger.warning('Everything is supposed to be closed')

    def stop(self) -> None:
        """Stop the execution."""
        self._alive = False

    def _append_events(self, events: Optional[EventOrEvents]) -> None:
        if events is None:
            return None
        elif isinstance(events, (list, tuple)):
            for event in reversed(events):
                if not self._is_terminal_event(event):
                    self._q.appendleft(event)
        elif not self._is_terminal_event(events):
            self._q.appendleft(events)

    def _is_terminal_event(self, event: Event) -> bool:
        if self._terminal_event is not None and isinstance(event, self._terminal_event):
            self.stop()
            return True
        return False

    def _close_pool(self) -> None:
        if self._pool is not None:
            self._pool.terminate()

    def _close_resources(self) -> None:
        self._close_pool()
        self._close_sources()
        self._close_processors()

    def _close_sources(self) -> None:
        for source in self._sources:
            source.close()

    def _close_processors(self) -> None:
        closed = set()
        for processors in self._processors.values():
            for processor in processors:
                identifier = id(processor)
                if identifier not in closed:
                    processor.close()
                    closed.add(identifier)

    def _get_new_events(self) -> None:
        for source in self._sources:
            self._append_events(source())

    def _handle_pool_exceptions(self, exc: BaseException, parallel_event: ParallelEvent):
        self.stop()
        logger.error('Execution failed during processing the event: {0}({1}) {2}'.format(
            parallel_event.function.__name__, parallel_event.trigger.__class__.__name__, parallel_event.trigger,
        ))
        logger.error(exc)

    def _process_events(self, input_event: Event) -> None:
        for processor in self._processors[type(input_event)]:
            try:
                output_event: Union[Event, ParallelEvent] = processor(input_event)
                if isinstance(output_event, ParallelEvent):
                    self._pool.apply_async(
                        run_parallel_processor,
                        (output_event.encode(), ),
                        callback=self._append_events,
                        error_callback=lambda exc: self._handle_pool_exceptions(exc, output_event)
                    )
                else:
                    self._append_events(output_event)
            except Exception:
                self.stop()
                logger.error('Execution failed during processing the event: {0}({1}) {2}'.format(
                    processor.__class__.__name__, input_event.__class__.__name__, input_event,
                ))
                logger.error(traceback.format_exc())

    def _tick(self) -> None:
        self._get_new_events()
        while self._q:
            event = self._q.popleft()
            self._process_events(event)


class AsyncRunner(object):
    """Same as Runner, but for async/await syntax."""

    def __init__(self, graph: Graph, terminal_event: Optional[Type[Event]]) -> None:
        """
        Init async runner instance.

        :param graph:           Computational graph to execute.
        :param terminal_event:  Special event which simply stops execution, when processed.
        """
        self._sources = graph.sources
        self._processors = graph.processors
        self._q: Deque[Event] = deque()
        self._alive = True
        self._terminal_event = terminal_event

    async def run(self, interval: float, once: bool) -> None:
        """
        Start execution.

        :param interval:        Minimal timeout (seconds) between repetition of the whole computational sequence.
        :param once:            If True, executes the whole computational sequence only once, otherwise won't stop until
                                specific conditions (such as terminal event) will occur.
        """
        await self._tick()
        if not once:
            while self._alive:
                start = time.perf_counter()
                await self._tick()
                delta = time.perf_counter() - start - interval
                if delta < 0:
                    await asyncio.sleep(abs(delta))
        await self._close_resources()

    def stop(self) -> None:
        """Stop the execution."""
        self._alive = False

    def _append_events(self, events: Optional[EventOrEvents]) -> None:
        if events is None:
            return None
        elif isinstance(events, (list, tuple)):
            for event in events:
                if not self._is_terminal_event(event):
                    self._q.append(event)
        elif not self._is_terminal_event(events):
            self._q.append(events)

    def _is_terminal_event(self, event: Event) -> bool:
        if self._terminal_event is not None and isinstance(event, self._terminal_event):
            self.stop()
            return True
        return False

    async def _close_resources(self) -> None:
        await self._close_sources()
        await self._close_processors()

    async def _close_processors(self) -> None:
        closed = set()
        for processors in self._processors.values():
            for processor in processors:
                identifier = id(processor)
                if identifier not in closed:
                    if isinstance(processor, AsyncProcessor):
                        await processor.close()
                    else:
                        processor.close()
                    closed.add(identifier)

    async def _close_sources(self) -> None:
        for source in self._sources:
            if isinstance(source, AsyncSource):
                await source.close()
            else:
                source.close()

    async def _get_new_events(self) -> None:
        tasks = []
        loop = asyncio.get_running_loop()
        for source in self._sources:
            if isinstance(source, AsyncSource):
                task: Awaitable[EventOrEvents] = loop.create_task(source())
            else:
                task = loop.run_in_executor(None, source)
            tasks.append(task)
        for events in await asyncio.gather(*tasks):
            self._append_events(events)

    async def _process_events(self) -> None:
        loop = asyncio.get_running_loop()
        tasks = []
        while self._q:
            event = self._q.popleft()
            for processor in self._processors[type(event)]:
                if isinstance(processor, AsyncProcessor):
                    task: Awaitable[EventOrEvents] = loop.create_task(processor(event))
                else:
                    prepared = functools.partial(processor, event)
                    task = loop.run_in_executor(None, prepared)
                tasks.append(task)

        for events in await asyncio.gather(*tasks):
            self._append_events(events)

        if self._q:
            await self._process_events()

    async def _tick(self) -> None:
        await self._get_new_events()
        await self._process_events()



"""
Blocks - *Sources* and *Processors* - are a backbone of you app.

Finally, all functions, decorated with :code:`@source` or :code:`@processor` are reborn to these classes.
"""
from enum import Enum
from abc import ABC, abstractmethod
from typing import Any, Union, Iterable, Optional, DefaultDict, Type, List


class TypeOfProcessor(Enum):
    SYNC = 0
    PARALLEL = 1


class Event(object):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        err = '. '.join([
            'You are going to use lightweight version of event',
            'Define __init__-method, or use dataclasses or pydantic instead.',
        ])
        raise NotImplementedError(err)


EventOrEvents = Union[Event, Iterable[Event]]


class Source(ABC):
    """
    Interface to represent single source of events, which emits them for the further processing.

    Class usage allows easily handle some internal state, if needed.

    Example::

      >>> from typing import NamedTuple, List

      >>> from blocks import Source

      >>> class MyEvent(NamedTuple):
      ...     ...

      >>> class MySource(Source):
      ...
      ...     def __init__(self, events: List[MyEvent]) -> None:
      ...         self.events = events
      ...
      ...     def __call__(self) -> List[MyEvent]:
      ...         return self.events[:]

      >>> blocks = (MySource(events=[MyEvent()]))
    """

    @abstractmethod
    def __call__(self) -> EventOrEvents:
        """
        Emit new events to the internal queue.

        :return:        Single event or sequence of events.
        """

    def close(self) -> None:
        """Define your graceful shutdown here."""


class Processor(ABC):
    """
    Interface to represent single unit of calculations, which reacts to some specific event.

    Class usage allows easily handle some state in processors, if needed.

    Example::

      >>> from typing import NamedTuple, Dict

      >>> from blocks import Processor

      >>> class MyEvent(NamedTuple):
      ...     ...

      >>> class Printer(Processor):
      ...
      ...     def __init__(self, state: Dict[str, int]) -> None:
      ...         self.state = state
      ...
      ...     def __call__(self, event: MyEvent) -> None:
      ...         print(event)

      >>> blocks = (Printer(state={}),)
    """

    _type_of_processor: int = TypeOfProcessor.SYNC

    @property
    def type_of_processor(self) -> int:
        return self._type_of_processor

    @type_of_processor.setter
    def type_of_processor(self, v: int) -> None:
        if v not in (i.value for i in TypeOfProcessor):
            raise ValueError("Incorrect type of processor")
        self._type_of_processor = v

    @abstractmethod
    def __call__(self, event: Any) -> Optional[EventOrEvents]:
        """
        Handle event.

        :param event:   Specific event to handle. If, for some reason, you want to process every event,
                        it is also possible by passing Any to implementation.
        :return:        Event or sequence of events to be processed later or by other processed.
                        Processor may return None, e.g. while implementing some effects in outer systems and
                        when result processing is unnecessary.
        """

    def close(self) -> None:
        """Define your graceful shutdown here."""


class AsyncSource(ABC):
    """
    Interface to represent single source of events, which emits them for the further processing.

    Class usage allows easily handle some internal state, if needed.

    Same as usual :code:`Source`, but awaitable.
    """

    @abstractmethod
    async def __call__(self) -> EventOrEvents:
        """
        Emit new events to the internal queue.

        :return:        Single event or sequence of events.
        """

    def close(self) -> None:
        """Define your graceful shutdown here."""


class AsyncProcessor(ABC):
    """
    Interface to represent single unit of calculations, which reacts to some specific event.

    Class usage allows easily handle some state in processors, if needed.

    Same as usual :code:`Processor`, but awaitable.
    """

    @abstractmethod
    async def __call__(self, event: Any) -> Optional[EventOrEvents]:
        """
        Handle event.

        :param event:   Specific event to handle. If, for some reason, you want to process every event,
                        it is also possible by passing Any to implementation.
        :return:        Event or sequence of events to be processed later or by other processed.
                        Processor may return None, e.g. while implementing some effects in outer systems and
                        when result processing is unnecessary.
        """

    def close(self) -> None:
        """Define your graceful shutdown here."""


AnySource = Union[AsyncSource, Source]
AnyProcessor = Union[AsyncProcessor, Processor]
Block = Union[AnySource, AnyProcessor]
AnyProcessors = DefaultDict[Type[Event], List[AnyProcessor]]



from blocks.kafka.app import KafkaApp
from blocks.kafka.events import Batch, CommitEvent, NoNewEvents
from blocks.kafka.topics import InputTopic, OutputTopic
from blocks.kafka.sources import KafkaSource
from blocks.kafka.processors import KafkaProducer, OffsetCommitter



from typing import Any, Dict, Callable, Optional
from datetime import datetime

# ToDo (tribunsky.kir): Implement something more flexible than {str: str}
Serializer = Callable[[Dict[str, Any]], Dict[bytes, bytes]]
Deserializer = Callable[[Dict[bytes, bytes]], Dict[str, Optional[bytes]]]


def deserialize(message: Dict[bytes, bytes]) -> Dict[str, Optional[bytes]]:
    dct = {}
    for key, value in message.items():
        to_insert = None if value == b'null' else value
        dct[key.decode()] = to_insert
    return dct


def serialize(dct: Dict[str, Any]) -> Dict[bytes, bytes]:
    message = {}
    for k, v in dct.items():
        if v is None:
            value = b'null'
        elif isinstance(v, bool):
            value = str(v).lower().encode()
        elif isinstance(v, bytes):
            value = v
        elif isinstance(v, datetime):
            value = str(int(v.timestamp() * 1000)).encode()
        else:
            value = str(v).encode()
        message[k.encode()] = value
    return message



from __future__ import annotations

import inspect
from typing import Any, List, Union, Callable

from blocks import Event, Source
from blocks.logger import logger
from blocks.db.types import Row, Query, Table
from blocks.postgres.protocols import Connection

# ToDo (tribunsky.kir): looks like we need some simple model-to-query derivation.
#                       Cause definition of model already dictates us which fields should be selected
#                       e.g. read_queries = [Query('select * from some_table', MyTableRow)]
#                       means like we want only specific columns from MyTable
#                       But it will restrict us to use simple queries only


def _get_rows_or_table(conn: Connection, query: Query) -> Union[List[Row], Table]:
    """Or how to make psycopg SLOW."""
    events = []
    if issubclass(query.codec, Row):
        row_codec = query.codec
    elif issubclass(query.codec, Table):
        # pydantic:
        # row_codec = query.codec.__annotations__['rows'].__args__[0]
        # dataclass:
        row_codec = query.codec.rows.__args__[0]
    else:
        raise TypeError('Not valid codec type: {0}'.format(type(query.codec)))
    with conn.cursor() as cur:
        logger.info(f'Executing query: {query.text}')
        cur.execute(query.text)
        logger.info('Validating')
        keys = [col.name for col in cur.description]
        fields = inspect.signature(row_codec).parameters
        fields_match = set(keys) == set(fields)
        if fields_match is False:
            logger.warning("Looks like query doesn't exactly matches model. Please, check.")
        for row in cur:
            row_dict = {key: value for key, value in zip(keys, row)}
            if fields_match is False:
                row_dict = {key: row_dict[key] for key in fields}
            events.append(row_codec(**row_dict))
    if issubclass(query.codec, Row):
        return events
    elif issubclass(query.codec, Table):
        return query.codec(rows=events)
    raise TypeError('Not valid codec type: {0}'.format(type(query.codec)))


class PostgresReader(Source):
    """
    Class represents event source that reads Postgres tables
    and wraps them into given events. On initialization must get list
    of queries in order to perform arbitrary mapping. In most cases
    you don't need to use PostgresReader directly, PostgresApp can make this for you.

    Example::

      >>> from dataclasses import dataclass
      >>>
      >>> from blocks import Graph
      >>> from blocks.postgres import PostgresReader, Query, Row

      >>> @dataclass
      ... class TableRow(Row):
      ...     x: int

      >>> queries = [Query('select * from some_table', TableRow)]
      >>> blocks = (PostgresReader(queries), ...)
    """
    # ToDo (tribunsky.kir): re-do on factory, which is able to create Connection on demand, or context manager.
    def __init__(self, queries: List[Query], conn: Callable[[], Connection]) -> None:
        self._conn_factory = conn
        self._queries = queries
        out_annots = {query.codec for query in queries}
        self.__call__.__annotations__['return'] = List[Union[tuple(out_annots)]]  # type: ignore

    def __call__(self) -> List[Event]:
        conn = self._conn_factory()
        try:
            with conn:
                rows_or_tables = self._run_queries(conn)
        finally:
            conn.close()
        return rows_or_tables

    def _run_queries(self, conn: Any) -> List[Event]:
        rows_or_tables: List[Event] = []
        for query in self._queries:
            row_or_table = _get_rows_or_table(conn, query)
            if isinstance(row_or_table, list):
                rows_or_tables.extend(row_or_table)
            else:
                rows_or_tables.append(row_or_table)
        return rows_or_tables



from typing import List, Type, Callable, Optional, Sequence

from blocks import App, Event, Runner, Processor
from blocks.db import Query
from blocks.validation import validate_blocks
from blocks.postgres.sources import PostgresReader
from blocks.postgres.protocols import Connection
from blocks.postgres.processors import PostgresWriter


class PostgresApp(App):
    """
    High level API for building applications
    with Postgres tables as inputs and possibly outputs.
    Wraps Graph creation and builds BatchApp as initialization.
    On initialization PostgresApp receives list of read queries,
    list of processors, list of update queries
    and save_graph flag with True as default.

    Example::

      >>> import psycopg2
      >>> from blocks import processor
      >>> from blocks.postgres import PostgresApp, Query, Row

      >>> class MyTableRow(Row):
      ...     x: int

      >>> @processor
      ... def printer(e: MyTableRow) -> None:
      ...     print(e)

      >>> def get_connection():
      ...     return psycopg2.connect(...)

      >>> read_queries = [Query('select * from some_table', MyTableRow)]
      >>> blocks = [printer()]
      >>> PostgresApp(get_connection, read_queries, blocks).run()
    """

    # ToDo (tribunsky.kir): define uniform API: queries in list which are separated in __init__
    def __init__(
        self,
        # ToDo (tribunsky.kir): leaky abstraction, check on more examples of libraries which implement PEP 249
        connection_factory: Callable[[], Connection],
        read_queries: Optional[List[Query]] = None,
        blocks: Optional[Sequence[Processor]] = None,
        update_queries: Optional[List[Query]] = None,
        terminal_event: Optional[Type[Event]] = None,
    ) -> None:
        super().__init__(blocks=[], terminal_event=terminal_event)
        if read_queries:
            self._graph.add_block(PostgresReader(read_queries, connection_factory))

        if blocks is not None:
            for processor in blocks:
                self._graph.add_block(processor)

        if update_queries:
            self._graph.add_block(PostgresWriter(update_queries, connection_factory))

        validate_blocks(self._graph.blocks)

    def run(self, *, min_interval: float = 0.0, once: bool = True) -> None:
        Runner(self._graph, self._terminal_event).run(interval=min_interval, once=once)



"""This module contains Kafka-specific processors."""

from typing import Any, Dict, List, Type, Union, Optional

from wunderkafka import AnyProducer, ProducerConfig, AvroModelProducer

from blocks.types import Event, Processor
from blocks.logger import logger
from blocks.kafka.types import ConsumersMapping
from blocks.kafka.events import CommitEvent
from blocks.kafka.topics import OutputTopic


def _assign_producers(
    topics: List[OutputTopic],
    config: Optional[ProducerConfig],
    cls: Type[AnyProducer],
) -> Dict[Type[Event], AnyProducer]:
    if len({topic.event for topic in topics}) != len(topics):
        raise ValueError('All events types should be unique!')

    instances: Dict[Type[AnyProducer], AnyProducer] = {}
    producers: Dict[Type[Event], AnyProducer] = {}

    for topic in topics:
        cfg = topic.producer.conf or config
        if cfg is None:
            raise ValueError('Consumer configuration must be set at least for all topic or per every topic.')
        producer_cls = topic.producer.type or cls
        instance = instances.get(producer_cls)
        if instance is None:
            instance = producer_cls(mapping=None, config=cfg)
            instances[producer_cls] = instance

        producers[topic.event] = instance
        instance.set_target_topic(topic.name, topic.event, topic.key)

    return producers


class KafkaProducer(Processor):
    """
    Class represents event processor that writes messages to Kafka topics
    based on data received from event. On initialization must get list
    of output topics in order to perform arbitrary mapping. In most cases
    you don't need to use KafkaProducer directly, KafkaApp can make this for you.

    Example::

      >>> from typing import NamedTuple

      >>> from blocks import Event, Graph
      >>> from blocks.kafka import KafkaProducer, OutputTopic

      >>> class MyEvent(NamedTuple):
      ...     x: int

      >>> topics = [OutputTopic('some_topic', MyEvent)]

      >>> blocks = (KafkaProducer(topics), ...)
    """

    def __init__(
        self,
        topics: List[OutputTopic],
        config: Optional[ProducerConfig],
        cls: Type[AnyProducer] = AvroModelProducer,
    ) -> None:
        """
        Init KafkaProducer instance.

        :param topics:      Topic to send messages in.
        :param config:      Failover configuration for producer if there is no configuration per InputTopic.
        :param cls:         Failover factory to create producer if there is no configuration per InputTopic.
        """
        self._producers: Dict[Type[Any], AnyProducer] = _assign_producers(topics, config, cls)
        self._topics: Dict[Type[Any], str] = {topic.event: topic.name for topic in topics}

        self.__call__.__annotations__.update(
            {
                'event': Union[tuple(topic.event for topic in topics)],
            },
        )

    def __call__(self, event: Event) -> None:
        """
        Send event-based message to corresponding topic.

        :param event:       Specific event to be serialized and sent to kafka topic.
        """
        key = vars(event).pop('@key', None)
        event_type = type(event)
        topic = self._topics[event_type]
        self._producers[event_type].send_message(topic, event, key)

    def close(self) -> None:
        """Graceful shutdown: flush all producers."""
        for producer in self._producers.values():
            # ToDo (tribunsky.kir): mimic/make public
            producer.flush()


# ToDo (dp.zubakin): handle corner cases, e.g.
#                    - Consistency: we may not commit offset due to the crash, but write result from this part of work.
#                                   So there may be problems with duplications in the next part of the pipeline
#                    - And vise versa, if we commit offset, but result sending failed.
class OffsetCommitter(Processor):
    """
    Class represents CommitEvent event processor that performs manual
    offset commit to list of given Kafka topics. On initialization must get
    refrence on dict of input topics to consumer mapping. In most cases
    you don't need to use OffsetCommitter directly, KafkaApp can make this for you.

    Example::

      >>> from blocks import Event
      >>> from blocks.kafka import KafkaSource, OffsetCommitter

      >>> topics = [...]
      >>> consumer = KafkaSource(topics)
      >>> committer = OffsetCommitter(consumer.consumers)

      >>> blocks = (consumer, committer, ...)
    """

    def __init__(self, consumers: ConsumersMapping) -> None:
        self._consumers = {topic.name: consumer for topic, consumer in consumers.items()}

    def __call__(self, event: CommitEvent) -> None:
        msg_metadata = vars(event.e).get('@msg')
        # ToDo (tribunsky.kir): thus, it is more reliable to put meta on EVERY message.
        #                       As we do not control user's CommitEvent usage, it may misbehave
        if msg_metadata is None:
            logger.warning('Got nothing to commit. Please check your commit_offset option for {0}'.format(event.e))
            return None
        to_commit = msg_metadata.to_commit()
        logger.debug('Really committing! {0} | {1}'.format(msg_metadata, to_commit))
        self._consumers[msg_metadata.topic].commit(offsets=to_commit)



"""This module contains specific per topic configuration definitions."""

import datetime
from enum import Enum
from typing import Type, Optional

from wunderkafka import AnyConsumer, AnyProducer, AvroConsumer, ConsumerConfig, ProducerConfig, AvroModelProducer

from blocks.types import Event
from blocks.logger import logger
from blocks.kafka.events import Batch


class ConsumerFactory(object):
    """Class to allow some narrow customization of consumers via InputTopic."""

    type: Type[AnyConsumer] = AvroConsumer
    conf: Optional[ConsumerConfig] = None


class ProducerFactory(object):
    """Class to allow some narrow customization of producers via OutputTopic."""

    type: Type[AnyProducer] = AvroModelProducer
    conf: Optional[ProducerConfig] = None


DEFAULT_CONSUMER = ConsumerFactory()
DEFAULT_PRODUCER = ProducerFactory()


class CommitChoices(Enum):
    """
    Enumeration represents allowed strategies for committing offsets of processed events via InputTopic.

    Options are:

        - 'never'
            Do not commit offset.
        - 'regular'
            Commit offset after the batch of messages has been consumed.
        - 'manual'
            For complicated cases when we want to commit offset only after the message has been actually processed.
            In that case, specific CommitEvent should be returned from specific processor.
    """

    never = 'never'
    regular = 'regular'
    manual = 'manual'


class _Topic(object):

    def __init__(
        self,
        name: str,
        event: Type[Event],
        *,
        key: Optional[Type[Event]] = None,
    ) -> None:
        self.name = name
        self.event = event
        self.key = key


class OutputTopic(_Topic):
    """
    Describes output event and specific producer's options for a given Kafka topic with a customizable key if necessary.

    Example::

      >>> from typing import NamedTuple

      >>> from blocks import Event
      >>> from blocks.kafka import KafkaProducer,

      >>> class MyEvent(NamedTuple):
      ...     x: int

      >>> topics = [OutputTopic('some_topic', MyEvent, debug='all')]
      >>> blocks = (KafkaProducer(topics), ...)
    """

    def __init__(
        self,
        name: str,
        event: Type[Event],
        *,
        key: Optional[Type[Event]] = None,
        producer: ProducerFactory = DEFAULT_PRODUCER,
    ) -> None:
        """
        Init Output's topic instance.

        :param name:            Kafka topic name to produce messages.
        :param event:           Event-inherited model to be serialized to kafka message's value.
        :param key:             Event-inherited model to be serialized to kafka message's key, if any.
        :param producer:        Factory which provides the way to instantiate producer.
        """
        super().__init__(name, event, key=key)
        self.producer = producer


class InputTopic(_Topic):
    """
    Describes input event stream from a given Kafka topic with a great flexibility.

    Example::

      >>> from typing import NamedTuple

      >>> from blocks import Event
      >>> from blocks.kafka import KafkaSource, InputTopic

      >>> class MyEvent(NamedTuple):
      ...     x: int

      >>> topics = [InputTopic('some_topic', MyEvent, from_beginning=True)]
      >>> blocks = (KafkaSource(topics), )
    """

    def __init__(  # noqa: WPS211  # ToDo (tribunsky.kir): InputTopic overloaded.
        self,
        name: str,
        event: Type[Event],
        *,
        group_id: Optional[str] = None,
        key: Optional[Type[Event]] = None,
        commit_offset: str = 'never',
        ignore_keys: bool = True,
        verbose_log_errors: bool = True,
        # per 'physical' consumer parameters, if needed.
        from_beginning: Optional[bool] = False,
        with_timedelta: Optional[datetime.timedelta] = None,
        poll_timeout: float = 0.05,
        messages_limit: int = 10000,
        # if any, send NoNewEvents
        dummy_events: bool = False,
        # it's all about batching.
        batch_event: Optional[Type[Batch]] = None,
        read_till: Optional[int] = None,
        max_empty_polls: int = 0,
        # allow to rebuild consumer
        consumer: ConsumerFactory = DEFAULT_CONSUMER,
    ) -> None:
        """
        Init InputTopic's instance.

        :param name:            Kafka topic name to be consumed.
        :param event:           Event-inherited model for message values to be unpacked to.
        :param group_id:        Override consumer's group id from config.
        :param key:             Event-inherited model for messages keys to be unpacked to, if any.
        :param commit_offset:   One of possible strategies of how to commit offsets for a given event.
        :param ignore_keys:     If True, don't try to deserialize message's keys. This flag defines behaviour of
                                underlying consumer, so if the event model isn't define for a given key and this flag
                                is not set to False, under the hood key will be deserialized.
        :param from_beginning:  If True, will read topic from the earliest offset, otherwise from the latest.
        :param with_timedelta:  Subscribe to topic via timestamp for a specific timedelta in the past.
        :param poll_timeout:    Interval to await for a new messages from Kafka broker.
        :param messages_limit:  Maximum amount of messages to be awaited for a given interval.
        :param dummy_events:    If True, emit NoNewEvents to the internal queue to notify processors, that there is no
                                new messages has been consumed for the last poll.
        :param batch_event:     If specified, packs all unpacked events to a single Batch event.
        :param read_till:       If set, will read topic until the given timestamp is reached. In addition to sequence
                                of events will also append built-in EndTsIsReached to notify producers.
        :param max_empty_polls: if set, will read topic until there are no new events consumed for a given count of
                                polls. If batch_event is specified, will pack messages to a single event, otherwise
                                behaves like read_till.
        :param consumer:        Factory which provides the way to instantiate consumer.

        :raises ValueError:     For broken invariants or incompatible options.
        """
        self.read_till = read_till
        self.max_empty_polls = max_empty_polls
        if self.max_empty_polls and self.read_till:
            exc_msg = 'Only one condition at time per topic is allowed, but both max_empty_polls and read_till are set'
            raise ValueError(exc_msg)
        if batch_event is not None and self.read_till is None:
            failover = int(abs(10 / poll_timeout)) or 1
            logger.warning('Using {0} as analog of 10 seconds for batch waiting failover.'.format(failover))
            self.max_empty_polls = failover
        if with_timedelta is not None and from_beginning is True:
            raise ValueError("Can't simultaneously read topic from the beginning and from specific timedelta")

        super().__init__(name, event, key=key)
        self.group_id = group_id
        self.from_beginning = from_beginning
        self.commit_offset = commit_offset
        self.ignore_keys = ignore_keys
        self.verbose_log_errors = verbose_log_errors
        self.with_timedelta = with_timedelta
        self.dummy_events = dummy_events
        self.batch_event = batch_event

        self.poll_timeout = poll_timeout
        self.messages_limit = messages_limit
        self.consumer = consumer

    # Shortening code in Consumers, but still exhibiting strings in API.
    @property
    def committable(self) -> bool:
        return CommitChoices(self.commit_offset) is not CommitChoices.never

    @property
    def commit_manually(self) -> bool:
        return CommitChoices(self.commit_offset) is CommitChoices.manual

    @property
    def commit_regularly(self) -> bool:
        return CommitChoices(self.commit_offset) is CommitChoices.regular

    @property
    def batched(self) -> bool:
        return self.batch_event is not None or self.read_till is not None



from blocks.types.base import (Event,
                               EventOrEvents,
                               Processor,
                               AsyncProcessor,
                               AnyProcessor,
                               Source,
                               AsyncSource,
                               AnySource,
                               Block,
                               TypeOfProcessor)
from blocks.types.dto import ParallelEvent



from typing import Any, Set, List, Type, Sequence

from blocks.types import Block, Event, Source, Processor, AsyncSource, AsyncProcessor
from blocks.logger import logger
from blocks.annotations import AnnotationError, _get_annotations, get_input_events_type, get_output_events_type

EXCLUDED_EVENTS = [type(None)]


def _flatten(list_of_lists: List[List[Any]]) -> List[Any]:
    return [val for sublist in list_of_lists for val in sublist]


def _output_events(blocks: Sequence[Block], excluded_events: Set[Type[Event]]) -> Set[Type[Event]]:
    nested_output_events = []
    for block in blocks:
        output_events = get_output_events_type(block)
        nested_output_events.append(output_events)
    return set(_flatten(nested_output_events)) - excluded_events


def _input_events(blocks: Sequence[Block]) -> Set[Type[Event]]:
    nested_input_events = []
    for block in blocks:
        if not isinstance(block, (AsyncSource, Source)):
            input_events = get_input_events_type(block)
            nested_input_events.append(input_events)
    return set(_flatten(nested_input_events))


def _useless_events(blocks: Sequence[Block], excluded_events: Set[Type[Event]]) -> Set[Type[Event]]:
    return _output_events(blocks, excluded_events) - _input_events(blocks)


def _useless_receivers(blocks: Sequence[Block], excluded_events: Set[Type[Event]]) -> Set[Block]:
    output_events_set = _output_events(blocks, excluded_events)
    useless_receivers: Set[Block] = set()
    for block in blocks:
        block_input_events = get_input_events_type(block)
        no_events_to_handle = all(event not in output_events_set for event in block_input_events)
        if not isinstance(block, (AsyncSource, Source)) and no_events_to_handle:
            useless_receivers.add(block)
    return useless_receivers


def _useless_producers(blocks: Sequence[Block], excluded_events: Set[Type[Event]]) -> Set[Block]:
    input_events_set = _input_events(blocks)
    useless_producers: Set[Block] = set()
    for ev in excluded_events:
        input_events_set.add(ev)
    for block in blocks:
        block_output_events = get_output_events_type(block)
        no_handlers_for_produced_events = all(event not in input_events_set for event in block_output_events)
        if block_output_events and no_handlers_for_produced_events:
            useless_producers.add(block)
    return useless_producers


def validate_blocks(blocks: Sequence[Block], excluded_events: Sequence[Type[Event]] = EXCLUDED_EVENTS) -> None:
    excl_set = set(excluded_events)
    for event in _useless_events(blocks, excl_set):
        logger.warning("{0} isn't processed".format(event))
    # ToDo (tribunsky.kir): bad validation, doesn't find Batches on output
    for processor in _useless_receivers(blocks, excl_set):
        logger.warning('{0} is useless. There is no events generated for'.format(processor))
    for processor in _useless_producers(blocks, excl_set):
        logger.warning('{0} is useless. There is no receivers for generated events'.format(processor))


def validate_annotations(block: Block) -> None:
    annotations = _get_annotations(block)
    if isinstance(block, (AsyncProcessor, Processor)):
        if len(annotations) < 2:
            raise AnnotationError(
                f"Given processor: {type(block).__name__} doesn't annotated well."
                'Make sure that input and output annotations properly written'
            )
        elif len(annotations) > 2:
            raise AnnotationError('Processor should have exactly one argument')

    elif isinstance(block, (AsyncSource, Source)):
        if len(annotations) != 1 and 'return' not in annotations:
            raise AnnotationError(
                f"Given source: {type(block).__name__} doesn't annotated well."
                'Make sure that output annotations properly written'
            )



from typing import List
from dataclasses import dataclass

from blocks.types import Event


@dataclass
class EndTsIsReached(Event):
    """
    Built-in event which notifies app that all messages since application start have been already received.

    Not supposed to be used directly.
    """

    topic_name: str


# ToDo (tribunsky.kir): limited usage ability. Maybe we need more tracing between blocks.
@dataclass
class CommitEvent(Event):
    """
    Built-in event which allows to commit offset if specific message has been already processed.

    Currently limited by events which are manually casted from kafka messages.

    Example::

      >>> from typing import Optional, NamedTuple
      >>> from blocks import Event, processor
      >>> from blocks.kafka import CommitEvent

      >>> class MyEvent(NamedTuple):
      ...     field1: int
      ...     field2: str

      >>> @processor
      >>> def proc(event: MyEvent) -> Optional[CommitEvent]:
      ...     ...
      ...     if event.field2 == 'some':
      ...         return CommitEvent(e=event)
    """

    e: Event  # noqa: WPS111  # ToDo (tribunsky.kir): awkward API. Do something with `e=` w/o overloading core event.


# ToDo (tribunsky.kir): IMHO, it is idiomatic enough to spawn such event, but there was an idea to get rid of it.
@dataclass
class NoNewEvents(Event):
    """
    Built-in event which allows to notify application, that there is no new messages polled from Kafka InputTopic.

    Example::

        >>> from blocks.kafka import NoNewEvents
        >>> from blocks import processor

        >>> @processor
        >>> def generator(event: NoNewEvents) -> None:
        ...     print('No new events from {event.source}')
    """

    source: str


@dataclass
class Batch(Event):
    """Built-in event which allows to send multiple messages to processors within single event."""

    events: List[Event]



import inspect
from typing import Any, Dict, List, Type, Tuple, Union, Optional, Sequence
from dataclasses import is_dataclass

from pydantic import ValidationError
from wunderkafka import Message, AnyConsumer, AvroConsumer, ConsumerConfig

from blocks.types import Event, Source
from blocks.logger import logger
from blocks.kafka.types import ConsumersMapping, KafkaMessageMeta
from blocks.kafka.events import NoNewEvents, EndTsIsReached
from blocks.kafka.topics import InputTopic


class KafkaSource(Source):
    """
    Class represents event source that reads messages from Kafka topics
    and wraps them into given events. On initialization must get list
    of input topics in order to perform arbitrary mapping. In most cases
    you don't need to use KafkaSource directly, KafkaApp can make this for you.

    Example::

      >>> from blocks import Event
      >>> from pydantic import BaseModel
      >>> from blocks.kafka import KafkaSource, InputTopic

      >>> class MyEvent(BaseModel):
      ...     x: int

      >>> topics = [InputTopic('some_topic', MyEvent)]
      >>> blocks = (KafkaSource(topics), )
    """

    def __init__(
        self,
        topics: List[InputTopic],
        config: Optional[ConsumerConfig] = None,
        *,
        cls: Type[AnyConsumer] = AvroConsumer,
        ignore_errors: bool = True,
    ) -> None:
        """
        Init KafkaSource instance.

        :param topics:          List of topics with corresponding settings to subscribe.
        :param config:          Failover configuration for consumer if there is no configuration per InputTopic.
        :param cls:             Failover factory to create consumer if there is no configuration per InputTopic.
        :param ignore_errors:   If True, do not fail during casting messages to events. By design, messages in Kafka may
                                be produced with totally different schemas, or without any.
        """
        self.consumers: ConsumersMapping = _init_consumers(topics, config, cls)
        self._commit_messages: Dict[InputTopic, Message] = {}
        self._ignore_errors = ignore_errors

        self._prev_poll: Dict[InputTopic, List[Message]] = {}
        out_annotations = {topic.event for topic in topics}
        # No magic, very explicit, aha.
        self.__call__.__annotations__['return'] = List[Union[tuple(out_annotations)]]  # type: ignore

    def __call__(self) -> List[Event]:
        """
        Emit messages from all topics as events to the internal queue.

        :return:        Single event or sequence of events.
        """
        self._commit()

        # If we've read some messages on previous iteration, we don't want to lose them just because they are too new
        events = []
        to_next: Dict[InputTopic, List[Message]] = {}

        for topic, consumer in self.consumers.items():
            if topic.batched or topic.max_empty_polls:
                messages, next_poll = _read_batch(consumer, topic)
            else:
                next_poll = []
                messages = consumer.consume(topic.poll_timeout, topic.messages_limit, ignore_keys=topic.ignore_keys)

            if topic.commit_regularly and messages:
                # Not using CommitEvent as we close enough to consumer itself. R. Reliability. C. Consistency.
                self._commit_messages[topic] = messages[-1]

            to_next[topic] = next_poll

            prev_events = self._prev_poll.get(topic, [])

            events.extend(_make_events(prev_events + messages, topic, self._ignore_errors))

        self._prev_poll = to_next

        return events

    def close(self) -> None:
        """Graceful shutdown: commit anything that should be committed as in InputTopic configurations."""
        self._commit()

    def _commit(self) -> None:
        for topic, message in self._commit_messages.items():
            self.consumers[topic].commit(message)
        if self._commit_messages:
            logger.debug(f'Messages: {self._commit_messages} committed')
            self._commit_messages.clear()


def _stash_msg_meta(event: Event, msg: Message) -> None:
    """Picking just necessary data to commit message later."""
    meta = KafkaMessageMeta(topic=msg.topic(), partition=msg.partition(), offset=msg.offset())
    event.__dict__['@msg'] = meta


def shortened(src: Dict[str, Any], n: int = 8) -> Dict[str, Any]:
    target = n - 3
    dct = {}
    for k, v in src.items():
        dct[k] = v
        if isinstance(v, (str, bytes)):
            dots = '...' if isinstance(v, str) else b'...'
            if len(v) > target:
                dct[k] = v[:target] + dots

    return dct


def cast(msg: Message, codec: Type[Event], *, ignore_errors: bool, verbose_log_errors: bool = True) -> Optional[Event]:
    # It actually works not only against instance, but against cls too
    message_value = msg.value()
    logger.debug('{0}|{1}|{2} (e: {3}): {4}'.format(
        msg.topic(), msg.partition(), msg.offset(), msg.error(), message_value),
    )
    if is_dataclass(codec):
        # ToDo (tribunsky.kir): move it to external cache OR
        #                       just do not use internally raw InputTopic's model on every message
        fields = inspect.signature(codec).parameters
        dct = {k: v for k, v in message_value.items() if k in fields}
    else:
        dct = message_value

    # ToDo (tribunsky.kir): definition of event via `= object` was a really 'nice' idea (no).
    try:
        return codec(**dct)
    except (ValidationError, TypeError) as e:
        if verbose_log_errors:
            logger.error(e)
        else:
            logger.error('Failed to extract the message ({0})'.format(shortened(dct)))
        if ignore_errors is False:
            raise
    return None


def _make_events(
    messages: List[Message],
    topic: InputTopic,
    ignore_errors: bool,
) -> List[Event]:
    events: List[Event] = []
    for msg in messages:
        event = cast(msg, topic.event, ignore_errors=ignore_errors, verbose_log_errors=topic.verbose_log_errors)
        if event is not None:
            if topic.commit_manually:
                # Do not know in advance which event should be committed.
                # So stashing necessary meta to every event from topics which may be committed manually.
                _stash_msg_meta(event, msg)
            events.append(event)

    if topic.batched:
        if topic.batch_event is None:
            events.append(EndTsIsReached(topic_name=topic.name))
        else:
            return [topic.batch_event(events=events)]
    else:
        if not events and topic.dummy_events:
            events.append(NoNewEvents(source=f'{topic.name} Kafka topic'))

    return events


def _init_consumers(
    topics: Sequence[InputTopic],
    config: Optional[ConsumerConfig],
    cls: Type[AnyConsumer],
) -> ConsumersMapping:
    # ToDo (tribunsky.kir): looks like gid should be set per topic.
    consumers = {}
    for topic in topics:
        if topic.group_id is None:
            if config is None:
                msg = ' '.join([
                    'Please, check group.id.',
                    'It may be set per InputTopic or should at least be provided via ConsumerConfig.',
                    'Currently both values are set to None.'
                ])
                raise ValueError(msg)
            else:
                gid = config.group_id
        else:
            gid = topic.group_id
        if topic.committable and gid is None:
            raise ValueError('For offset committing group_id must be provided')

        cfg = topic.consumer.conf or config
        if cfg is None:
            raise ValueError('Consumer configuration must be set at least for all topic or per every topic.')
        if topic.group_id:
            cfg.group_id = topic.group_id

        cns = topic.consumer.type or cls
        consumer = cns(cfg)

        if topic.committable:
            consumer.subscribe([topic.name])
        else:
            if topic.with_timedelta is not None:
                consumer.subscribe([topic.name], with_timedelta=topic.with_timedelta)
            else:
                consumer.subscribe([topic.name], from_beginning=topic.from_beginning)
        consumers[topic] = consumer
    return consumers


def _read_batch(consumer: AnyConsumer, topic: InputTopic) -> Tuple[List[Message], List[Message]]:
    if topic.read_till is not None:
        return _read_topic_till_ts(consumer, topic)
    if topic.max_empty_polls > 0:
        return _read_topic_till_end(consumer, topic), []
    msg_template = 'Wrong read_till ({0}) or max_empty_polls ({1}) for {2}'
    logger.warning(msg_template.format(topic.read_till, topic.max_empty_polls, topic.name))
    return [], []


def _read_topic_till_ts(consumer: AnyConsumer, topic: InputTopic) -> Tuple[List[Message], List[Message]]:
    # Search every topic partition till any messages with appropriate ts are polled
    messages = []
    also_polled = []
    should_read = True
    while should_read:
        msgs = consumer.consume(topic.poll_timeout, topic.messages_limit, ignore_keys=topic.ignore_keys)
        if msgs:
            appended = False
            for msg in msgs:
                _, ts = msg.timestamp()
                if ts <= topic.read_till:
                    messages.append(msg)
                    appended = True
                else:
                    also_polled.append(msg)
            should_read = appended is True and len(messages) > 0
    return messages, also_polled


def _read_topic_till_end(consumer: AnyConsumer, topic: InputTopic) -> List[Message]:
    # It may be implemented via other strategy: discover maximum offset in topic via subscription to OFFSET_END
    messages = []
    empty_polls_count = 0
    while empty_polls_count <= topic.max_empty_polls:
        msgs = consumer.consume(topic.poll_timeout, topic.messages_limit, ignore_keys=topic.ignore_keys)
        if msgs:
            messages.extend(msgs)
            empty_polls_count = 0
        else:
            empty_polls_count += 1
    return messages



"""Graph - is just static collection of blocks, which is waiting a moment to be executed."""

from typing import Set, List, Type, Optional, Sequence, Union
from collections import defaultdict

from blocks.graph_utils import build_graph
from blocks.logger import logger

from blocks.types import Block, Event, Source, AnySource, Processor, AsyncSource, AsyncProcessor
from blocks.types.base import AnyProcessors, TypeOfProcessor
from blocks.types.graph import RenderingKernelType
from blocks.validation import validate_annotations
from blocks.annotations import get_input_events_type, get_output_events_type


class Graph(object):
    """Represents computational graph, consisting from the given blocks and events."""

    def __init__(self, blocks: Optional[Sequence[Block]] = None) -> None:
        """
        Init Graph instance.

        Splits all given blocks into two groups: sources and processors.

        :param blocks:      Sequence of blocks to be added to computational graph.
        """
        self.sources: List[AnySource] = []
        self.processors: AnyProcessors = defaultdict(list)
        self.count_of_parallel_tasks: int = 0

        self._output_events: Set[Type[Event]] = set()

        if blocks is not None:
            for block in blocks:
                self.add_block(block)

    def add_block(self, block: Block) -> None:
        """
        Add block to current graph.

        :param block:       Processor or Source to be included in graph.
        """
        validate_annotations(block)
        if isinstance(block, Processor) and block.type_of_processor == TypeOfProcessor.PARALLEL:
            self.count_of_parallel_tasks += 1
        if isinstance(block, (AsyncSource, Source)):
            self.sources.append(block)
        elif isinstance(block, (AsyncProcessor, Processor)):
            input_events = get_input_events_type(block)
            for event in input_events:
                self.processors[event].append(block)

            for event in get_output_events_type(block):
                self._output_events.add(event)

        else:
            raise TypeError(
                'Wrong block type: {0}, should be AsyncSource, Source, AsyncProcessor or Processor subclass'.format(
                    type(block),
                ),
            )

    def save_visualization(
        self,
        rendering_type: Union[str, RenderingKernelType] = RenderingKernelType.matplotlib
    ) -> None:
        """
        Build graph by using NetworkX and render by matplotlib
        saves default 'GRAPH.png' visualization on disk in the root folder
        or
        Build graphviz Digraph self representation
        and if renderer installed in the system saves
        'DAG.svg' visualization on disk in the root folder

        :param rendering_type: string or Enum type
            string only one of ['matplotlib', 'graphviz']
        """
        _types = [RenderingKernelType.matplotlib, RenderingKernelType.graphviz]
        if rendering_type not in _types:
            raise ValueError('Bad value for the argument, the value must be one of [matplotlib, graphviz]')
        try:
            build_graph(self.blocks, self.processors, rendering_type)
        except RuntimeError as ex:
            raise ex
        except Exception as ex:
            logger.error(ex)

    @property
    def contains_async_blocks(self) -> bool:
        for source in self.sources:
            if isinstance(source, AsyncSource):
                return True
        for processors in self.processors.values():
            for processor in processors:
                if isinstance(processor, AsyncProcessor):
                    return True
        return False

    @property
    def outputs(self) -> Set[Type[Event]]:
        return self._output_events

    @property
    def blocks(self) -> List[Block]:
        list_of_lists = [lst for lst in self.processors.values()]
        processors: List[Block] = [item for sublist in list_of_lists for item in sublist]
        sources: List[Block] = list(self.sources)
        blocks = processors + sources
        return blocks



from dataclasses import dataclass
from typing import Callable, Optional

from blocks.types.base import Event, EventOrEvents
try:
    import dill
except ImportError:
    dill = None


@dataclass
class ParallelEvent(Event):
    function: Callable[[Event], Optional[EventOrEvents]]
    trigger: Event

    def encode(self) -> bytes:
        if dill is None:
            raise RuntimeError("For parallel events needed dill library. Please install it")
        return dill.dumps((self.function, self.trigger))

    @classmethod
    def decode(cls, payload: bytes) -> "ParallelEvent":
        if dill is None:
            raise RuntimeError("For parallel events needed dill library. Please install it")
        function, trigger = dill.loads(payload)
        return cls(function=function, trigger=trigger)



from typing import List, Union, Callable

from psycopg2 import sql

from blocks import Processor
from blocks.logger import logger
from blocks.db.types import Row, Query, Table
from blocks.postgres.protocols import Connection


# ToDo (tribunsky.kir): not sure if it is good idea to give the user execute any queries
def _exec_queries(conn: Connection, query_text: str, rows: Table) -> None:
    columns = rows.columns
    query = sql.SQL(query_text).format(
        sql.SQL(', ').join([sql.Identifier(col) for col in columns]),
        sql.SQL(', ').join(sql.Placeholder() * len(columns)),
    )
    with conn.cursor() as cursor:
        insert_tuples = rows.values
        # ToDo (tribunsky.kir): optimize; executemany is slow too https://github.com/psycopg/psycopg2/issues/491
        #                       Maybe a nice approach would be: psycopg2.extras.execute_values
        total_count = len(insert_tuples)
        query_formatted = query.as_string(conn)
        for ix, tpl in enumerate(insert_tuples, 1):
            logger.info('[{0}/{1}] Executing query: {2} {3}'.format(ix, total_count, query_formatted, tpl))
            cursor.execute(query, tpl)


class PostgresWriter(Processor):
    """
    Class represents event processor that writes messages to Postgres tables
    based on data received from event. On initialization must get list
    of queries in order to perform arbitrary mapping. In most cases
    you don't need to use PostgresReader directly, PostgresApp can make this for you.

    Example::

      >>> from dataclasses import dataclass
      >>>
      >>> from blocks import Graph
      >>> from blocks.postgres import PostgresWriter, Query, Row

      >>> @dataclass
      ... class TableRow(Row):
      ...     x: int

      >>> queries = [Query('insert into table ({}) values ({})', TableRow)]
      >>> blocks = (PostgresWriter(queries), ...)
    """

    def __init__(self, queries: List[Query], connection_factory: Callable[[], Connection]) -> None:
        self._connection_factory = connection_factory
        self._conn = self._connection_factory()
        self._queries = {query.codec: query for query in queries}
        self._closed = False
        self.__call__.__annotations__.update(
            {
                'event': Union[tuple(query.codec for query in queries)],
            },
        )

    def __call__(self, event: Union[Row, Table]) -> None:
        with self._conn:
            self._run_query(self._conn, event)

    def _run_query(self, conn: Connection, event: Union[Row, Table]) -> None:
        query = self._queries[type(event)]
        if isinstance(event, Row):
            table = Table(rows=[event])
        elif isinstance(event, Table):
            table = event
        else:
            raise ValueError('Wrong event type {0}'.format(event))
        _exec_queries(conn, query.text, table)

    def close(self) -> None:
        if not self._closed:
            self._conn.close()
            self._closed = True



import inspect
from sqlite3 import Connection

from typing import List, Union, Callable, Any

from blocks import Source, Event
from blocks.logger import logger
from blocks.db.types import Row, Table
from blocks.db.next.sql import Query


def _get_rows_or_table(conn: Connection, query: Query) -> Union[List[Row], Table]:
    """Or how to make psycopg SLOW."""
    events = []
    if issubclass(query.codec, Row):
        row_codec = query.codec
    elif issubclass(query.codec, Table):
        # pydantic:
        # row_codec = query.codec.__annotations__['rows'].__args__[0]
        # dataclass:
        row_codec = query.codec.rows.__args__[0]
    else:
        raise TypeError('Not valid codec type: {0}'.format(type(query.codec)))
    # ToDo (tribunsky.kir): sqlite raises `AttributeError: __enter__`
    # with conn.cursor() as cur:
    cur = conn.cursor()
    logger.info(f'Executing query: {query.text}')
    cur.execute(query.text)
    logger.info('Validating')
    # ToDo (tribunsky.kir): sqlite returns only tuple with names
    # keys = [col.name for col in cur.description]
    keys = [col[0] for col in cur.description]
    fields = inspect.signature(row_codec).parameters
    fields_match = set(keys) == set(fields)
    if fields_match is False:
        logger.warning("Looks like query doesn't exactly matches model. Please, check.")
    for row in cur:
        row_dict = {key: value for key, value in zip(keys, row)}
        if fields_match is False:
            row_dict = {key: row_dict[key] for key in fields}
        events.append(row_codec(**row_dict))
    if issubclass(query.codec, Row):
        return events
    elif issubclass(query.codec, Table):
        return query.codec(rows=events)
    raise TypeError('Not valid codec type: {0}'.format(type(query.codec)))


class SQLReader(Source):

    def __init__(self, queries: List[Query], conn: Callable[[], Connection]) -> None:
        self._conn_factory = conn
        self._queries = queries
        out_annots = {query.codec for query in queries}
        self.__call__.__annotations__['return'] = List[Union[tuple(out_annots)]]  # type: ignore

    def __call__(self) -> List[Event]:
        conn = self._conn_factory()
        try:
            # with conn:
            rows_or_tables = self._run_queries(conn)
        finally:
            conn.close()
        return rows_or_tables

    def _run_queries(self, conn: Any) -> List[Event]:
        rows_or_tables: List[Event] = []
        for query in self._queries:
            row_or_table = _get_rows_or_table(conn, query)
            if isinstance(row_or_table, list):
                rows_or_tables.extend(row_or_table)
            else:
                rows_or_tables.append(row_or_table)
        return rows_or_tables




from typing import Type, Union, Optional, Sequence

from redis import Redis

from blocks import App, Block, Event
from blocks.validation import validate_blocks
from blocks.redis.serdes import Serializer, Deserializer, serialize, deserialize
from blocks.redis.sources import RedisConsumer
from blocks.redis.streams import InputStream, OutputStream
from blocks.redis.processors import RedisProducer


# ToDo (tribunsky.kir): from now it looks like GREAT copypaste of KafkaApp,
#                       so, maybe it makes sense to start using metaclasses here
class RedisStreamsApp(App):
    """
    High level API for building applications with redis streams as inputs and outputs.

    Wraps Graph creation and builds RedisStreamsApp.

    Example::

      >>> from typing import NamedTuple

      >>> from redis import Redis
      >>> from blocks import Event, processor
      >>> from blocks.redis import RedisStreamsApp, InputStream

      >>> class MyEvent(NamedTuple):
      ...     x: int

      >>> @processor
      ... def printer(e: MyEvent) -> None:
      ...     print(e)

      >>> streams = [InputStream('some_stream', MyEvent)]
      >>> blocks = [printer()]
      >>> redis = Redis(host=..., port=..., db=..., password=...)
      >>> RedisStreamsApp(streams, blocks, redis).run()
    """
    # ToDo (tribunsky.kir): maybe it is worth to add serdes per stream.
    def __init__(
        self,
        streams: Sequence[Union[InputStream, OutputStream]],
        blocks: Sequence[Block],
        redis_client: Redis,
        read_timeout: int = 100,
        terminal_event: Optional[Type[Event]] = None,
        serializer: Serializer = serialize,
        deserializer: Deserializer = deserialize,
    ) -> None:
        super().__init__(blocks=[], terminal_event=terminal_event)
        for block in blocks:
            self._graph.add_block(block)

        input_streams = [stream for stream in streams if isinstance(stream, InputStream)]
        output_streams = [stream for stream in streams if isinstance(stream, OutputStream)]

        if input_streams:
            self._graph.add_block(RedisConsumer(redis_client, input_streams, read_timeout, deserializer=deserializer))

        if output_streams:
            self._graph.add_block(RedisProducer(redis_client, output_streams, serializer=serializer))

        validate_blocks(self._graph.blocks)



from __future__ import annotations

from typing import Any, List, Type, Union
from datetime import datetime

from schedule import Job

from blocks.types import Event, Source, EventOrEvents


class Scheduler(Source):
    """
    Class represents source of events which are emitted by schedule.

    Therefore, other processors in your program may use these events as time-driven triggers.

    Uses "schedule" library (https://schedule.readthedocs.io/en/stable/)
    under the hood and exposes only subset of it's API.

    Example::

      >>> from blocks import Event
      >>> from blocks.sources.schedule import Scheduler

      >>> class MyEvent(Event):
      ...     pass

      >>> blocks = (Scheduler(MyEvent).every().minute.at(':30'), ...)
    """

    def __init__(self, scheduled_event: Type[Event]) -> None:
        self._event = scheduled_event
        self._job = Job(1)
        self.__call__.__annotations__['return'] = List[scheduled_event]  # type: ignore

    def __call__(self, _: Any = None) -> EventOrEvents:
        if self._job.should_run:
            self._job.last_run = datetime.now()
            self._job._schedule_next_run()
            return [self._event()]
        return []

    def every(self, interval: Union[int, float] = 1) -> Scheduler:
        self._job.interval = interval
        return self

    @property
    def second(self) -> Scheduler:
        return self.__hide_schedule_api('seconds')

    @property
    def minute(self) -> Scheduler:
        return self.__hide_schedule_api('minutes')

    @property
    def hour(self) -> Scheduler:
        return self.__hide_schedule_api('hours')

    @property
    def seconds(self) -> Scheduler:
        return self.second

    @property
    def minutes(self) -> Scheduler:
        return self.minute

    @property
    def hours(self) -> Scheduler:
        return self.hour

    def at(self, time_str: str) -> Scheduler:
        self._job.at(time_str)
        self._job._schedule_next_run()
        return self

    def __hide_schedule_api(self, unit: str) -> Scheduler:
        self._job.unit = unit
        self._job._schedule_next_run()
        return self



"""This module contains specific event types for Kafka."""

from typing import Dict, List, Optional, NamedTuple

from wunderkafka import AnyConsumer
from confluent_kafka import TopicPartition

from blocks.logger import logger
from blocks.kafka.topics import InputTopic

# ToDo (tribunsky.kir): make immutable?
ConsumersMapping = Dict[InputTopic, AnyConsumer]


class KafkaMessageMeta(NamedTuple):
    """
    Meta required for topic offset commit. Not supposed to be used directly.

    Optionals from original API:
    https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#message
    """

    topic: Optional[str]
    partition: Optional[int]
    offset: Optional[int]

    def to_commit(self) -> List[TopicPartition]:
        """
        As we do not use binary Message itself, we should increase offset by ourselves.

        - message (confluent_kafka.Message) – Commit message’s offset+1.
        - offsets (list(TopicPartition)) – List of topic+partitions+offsets to commit.

        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#confluent_kafka.Consumer.commit

        :return:    List of TopicPartitions to be committed by real consumer.
        """
        if self.topic is not None and self.partition is not None and self.offset is not None:
            tpn = TopicPartition(
                topic=self.topic,
                partition=self.partition,
                offset=self.offset + 1,
            )
            return [tpn]
        logger.warning('Some of required to commit data is absent: topic={0}, partition={1}, offset={2}'.format(
            self.topic,
            self.partition,
            self.offset,
        ))
        return []



from __future__ import annotations

from typing import List, Type, Tuple, Union
from dataclasses import asdict, dataclass

from blocks import Event


@dataclass
class Row(Event):
    """
    Base class for database row description.
    Performs runtime type checks on table row received
    from query and drops invalid rows. Database source will returns
    every row as a single event in this case.
    Example::

      >>> from typing import Optional
      >>>
      >>> from blocks.db import Row
      >>>
      >>> class TableRow(Row):
      ...     field1: Optional[int]
      ...     field2: str
      >>>
      >>> query = Query('SELECT * FROM some_table', TableRow)
    """
    ...

    @property
    def columns(self) -> List[str]:
        return list(asdict(self).keys())

    @property
    def values(self) -> Tuple:
        return tuple(asdict(self).values())


@dataclass
class Table(Event):
    """
    Base class for database table description.
    Performs runtime type checks on every received row
    from query and drops invalid rows and then groups
    rows into single table. Database source will returns
    all rows grouped in list as a single event in this case.
    Example::

      >>> from typing import Optional
      >>> from blocks.db import Row, Table
      >>>
      >>> class SomeTableRow(Row):
      ...     field1: Optional[int]
      ...     field2: str
      >>>
      >>> class SomeTable(Table):
      ...     rows: List[SomeTableRow]
      >>>
      >>> query = Query('SELECT * FROM some_table', SomeTable)
    """
    rows: List[Row]

    @property
    def columns(self) -> List[str]:
        # ToDo (tribunsky.kir): stupid, but will work for simple cases
        if self.rows:
            return self.rows[0].columns
        return []

    @property
    def values(self) -> List[Tuple]:
        return [row.values for row in self.rows]


@dataclass
class Query(Event):
    """
    Database query class that maps query result to given rows or table classes
    in case of selection and fills query with data from given row
    in case of insertion or updating.

    Example::

      >>> from typing import Optional
      >>> from blocks.db import Row, Table
      >>>
      >>> class SomeTableRow(Row):
      ...     field1: Optional[int]
      ...     field2: str
      >>>
      >>> class SomeTable(Table):
      ...     rows: List[SomeTableRow]
      >>>
      >>> query1 = Query('SELECT * FROM some_table', SomeTable)
      >>> query2 = Query('SELECT * FROM some_table', SomeTableRow)
    """
    text: str
    codec: Type[Union[Row, Table]]



from blocks.redis.app import RedisStreamsApp
from blocks.redis.sources import RedisConsumer
from blocks.redis.streams import InputStream, OutputStream, _Stream
from blocks.redis.processors import RedisProducer
